import cv2
import matplotlib.pyplot as plt
import numpy as np


def get_bounding_box_of_mask(mask):
    """
    Args:
        mask (2D numpy array): the value of mask is 0 and 1

    Returns:
        top_left_point (numpy array with shape = (2,)): coordinate of bbox
        bottom_right_point (numpy array with shape = (2,)): coordinate of bbox
    """
    nonzero_points = cv2.findNonZero(mask)
    x, y, width, height = cv2.boundingRect(nonzero_points)

    return (np.array([x, y]), np.array([x + width, y + height]))


def show_mask(mask, figsize=None):
    if figsize:
        plt.figure(figsize=figsize)
    plt.imshow(mask, cmap="gray")
    plt.show()



import random
import shutil
from pathlib import Path

import cv2
import numpy as np
import yaml

from kano.detect_utils import draw_bbox, xywh2xyxy
from kano.file_utils import create_folder, list_files
from kano.image_utils import concatenate_images, show_image
from kano.pose_utils import draw_skeleton

TASKS = ["detect", "pose"]


class YoloImage:
    """
    Represents an image annotated in YOLO format, which includes bounding boxes or skeletons.

    Attributes:
        image (numpy.ndarray): The original image.
        image_path (str): Path to the image file.
        label_path (str): Path to the label file corresponding to the image.
        task (str): Task type, either "detect" or "pose".
        labels (list): List of dictionaries containing label information.
        labels_dict (dict): Dictionary mapping class IDs to class names.
    """

    def __init__(self, image_path, labels_dict=None, task="detect"):
        """
        Initialize a YoloImage object.

        Args:
            image_path (str): Path to the image file.
            labels_dict (dict): Dictionary mapping class IDs to class names.
            task (str): Task type. Possible values: "detect", "pose".
        """
        self.image = cv2.imread(image_path)
        self.image_path = image_path
        self.label_path = self.get_label_path(image_path)
        if task not in TASKS:
            raise ValueError("Unexpected task. Please provide one of:", TASKS)
        self.task = task
        self.labels = self.get_labels(self.label_path)
        self.labels_dict = labels_dict

    def get_label_path(self, image_path):
        """
        Get the path to the label file corresponding to the given image.

        Args:
            image_path (str): Path to the image file.

        Returns:
            label_path (str): Path to the label file.
        """
        image_path = Path(image_path)
        images_folder_path = image_path.parent
        dataset_path = images_folder_path.parent
        labels_folder_path = dataset_path / "labels"
        label_filename = image_path.with_suffix(".txt").name
        label_path = labels_folder_path / label_filename

        return str(label_path)

    def get_labels(self, label_path):
        """
        Parse the label file and extract label information.
        Each label can contain:
            - class: class of the object
            - s_xywh: scaled xywh box
            - xyxy: xyxy box
            - keypoints: list of dict(xy, state) for pose estimation tasks

        Args:
            label_path (str): Path to the label file.

        Returns:
            labels (list): List of dictionaries, each containing label information.
        """
        labels = list()
        image_height, image_width = self.image.shape[:2]
        with open(label_path, "r") as file:
            for line in file:
                line = line.strip().split()
                label = {
                    "class": int(line[0]),
                    "s_xywh": np.array([float(x) for x in line[1:5]]),
                }
                xywh = label["s_xywh"].copy() * np.array(
                    [image_width, image_height, image_width, image_height]
                )
                label["xyxy"] = xywh2xyxy(xywh)

                if self.task == "pose":
                    keypoints = list()
                    for start_id in range(5, len(line), 3):
                        # "state" in [0, 1, 2] with:
                        # - 0: deleted
                        # - 1: occluded
                        # - 2: visible
                        x = int(float(line[start_id]) * image_width)
                        y = int(float(line[start_id + 1]) * image_height)
                        state = int(float(line[start_id + 2]))
                        keypoints.append(
                            {
                                "xy": (x, y),
                                "state": state,
                            }
                        )
                    label["keypoints"] = keypoints

                labels.append(label)
        return labels

    def show_image(self, figsize=(10, 10)):
        """
        Display the original image.

        Args:
            figsize (tuple(int, int)): Size of the figure (width, height) in inches.
        """
        show_image(self.image, figsize)

    def get_annotated_image(self):
        """
        Get the annotated image with bounding boxes or skeletons drawn.

        Returns:
            annotated_image (np.ndarray): Annotated image.
        """
        annotated_image = self.image.copy()
        for label in self.labels:

            if self.task == "pose":
                annotated_image = draw_skeleton(
                    annotated_image, label["keypoints"]
                )

            cls = label["class"]
            if self.labels_dict is not None:
                cls = self.labels_dict[cls]
            bbox = label["s_xywh"]

            annotated_image = draw_bbox(
                annotated_image, bbox, "s_xywh", (0, 255, 0), str(cls)
            )

        return annotated_image

    def show_annotated_image(self, figsize=(10, 10)):
        """
        Display the annotated image with bounding boxes or skeletons drawn.

        Returns:
            figsize (tuple(int, int)): Size of the figure (width, height) in inches.
        """
        annotated_image = self.get_annotated_image()
        show_image(annotated_image, figsize)

    def copy_to(
        self,
        target_folder_path,
        prefix="",
        reindex_dict=None,
        target_stem=None,
    ):
        """
        Copy the image and its label file to the specified target folder.

        Returns:
            target_folder_path (str): Path to the target folder.
            prefix (str): Prefix to be added to the filenames
            reindex_dict (dict): Dictionary mapping original class IDs to new class IDs.
            target_stem (str): Name stem for the copied files
        """
        source_image_path = Path(self.image_path)
        source_label_path = Path(self.label_path)
        target_folder_path = Path(target_folder_path)

        if target_stem is None:
            target_image_path = (
                target_folder_path
                / "images"
                / (prefix + source_image_path.name)
            )
            target_label_path = (
                target_folder_path
                / "labels"
                / (prefix + source_label_path.name)
            )
        else:
            target_image_path = (
                target_folder_path / "images" / (target_stem + ".jpg")
            )
            target_label_path = (
                target_folder_path / "labels" / (target_stem + ".txt")
            )

        shutil.copyfile(self.image_path, str(target_image_path))
        shutil.copyfile(self.label_path, str(target_label_path))

        if reindex_dict:
            with open(str(source_label_path), "r") as f:
                lines = f.readlines()

            new_lines = list()
            for line in lines:
                class_id = int(line.split()[0])
                if class_id in reindex_dict:
                    new_class_id = reindex_dict[class_id]
                    if new_class_id is not None:
                        new_line = (
                            f"{new_class_id} {' '.join(line.split()[1:])}\n"
                        )
                        new_lines.append(new_line)

            with open(str(target_label_path), "w") as f:
                f.writelines(new_lines)


class YoloDataset:
    """
    Dataset class with Yolo format.

    Attributes:
        dataset_path (Path): path to the dataset folder
        name (str): name of the dataset (folder name)
        classes (list(str)): names of classes in the dataset
        task (str): dataset usecase, must be a task in ["detect", "pose"]
    """

    def __init__(self, dataset_path, task="detect"):
        """
        Initialize a YoloDataset object.

        Returns:
            dataset_path (str): Path to the dataset folder.
            task (str): . Dataset use case. Default is "detect".
                        Possible values: "detect", "pose".
        """
        self.dataset_path = Path(dataset_path)
        self.name = self.dataset_path.name
        self.train_folder = self.dataset_path / "train"
        self.valid_folder = self.dataset_path / "valid"
        self.test_folder = self.dataset_path / "test"
        self.classes = self.get_classes(str(self.dataset_path / "data.yaml"))
        if task not in TASKS:
            raise ValueError("Unexpected task. Please provide one of:", TASKS)
        self.task = task

    @classmethod
    def get_classes(cls, yaml_path):
        """
        Retrieve classes from a YAML file.

        Returns:
            yaml_path (str): Path to the YAML file containing class names.

        Returns:
            names (list(str)): List of class names.
        """
        with open(yaml_path, "r") as file:
            data = yaml.safe_load(file)
        return data["names"]

    def summary(self):
        """
        Print summary information about the dataset.
        """
        print(f"Summary dataset {self.dataset_path.name}:")
        print("- Classes: ", self.classes)
        print("- Subsets:")
        total_file_count = 0
        for folder_path in [
            self.train_folder,
            self.valid_folder,
            self.test_folder,
        ]:
            if folder_path.exists():
                file_count = len(list_files(str(folder_path / "images")))
                print(f"  + {folder_path.name}: {file_count} images")
                total_file_count += file_count
        print("- Total images:", total_file_count)

    @classmethod
    def _combine_classes(cls, datasets_paths):
        """
        Combine classes from multiple datasets.

        Returns:
            datasets_paths (list(str)): Paths to the dataset folders.

        Returns:
            classes (list(str)): Combined list of class names.
        """
        classes = set()

        for path in datasets_paths:
            new_classes = cls.get_classes(str(Path(path) / "data.yaml"))
            classes.update(new_classes)

        classes = list(classes)
        classes.sort()

        return classes

    @classmethod
    def _get_reindex_dict(cls, source_classes, target_classes):
        """
        Get the reindexing dictionary for mapping classes from source to target.

        Returns:
            source_classes (list(str)): Source class names.
            target_classes (list(str)): Target class names.

        Returns:
            reindex_dict (dict): Reindexing dictionary.
        """
        reindex_dict = dict()
        for i, class_name in enumerate(source_classes):
            if class_name in target_classes:
                reindex_dict[i] = target_classes.index(class_name)

        return reindex_dict

    @classmethod
    def _create_simple_yaml_file(cls, dataset_path, classes):
        """
        Create a simple YAML file containing dataset information.

        Returns:
            dataset_path (str): Path to the dataset folder.
            classes (list[str]): List of class names.
        """
        data = {
            "train": "train",
            "val": "valid",
            "test": "test",
            "names": classes,
        }

        yaml_path = Path(dataset_path) / "data.yaml"
        with open(str(yaml_path), "w") as f:
            yaml.dump(data, f)

    @classmethod
    def merge_datasets(cls, datasets_paths, merged_dataset_path):
        """
        Merge multiple datasets into one.

        Returns:
            datasets_paths (list[str]): Paths to the dataset folders to be merged.
            merged_dataset_path (str): Path to the merged dataset folder.
        """
        merged_dataset_path = Path(merged_dataset_path)
        merged_classes = cls._combine_classes(datasets_paths)
        print("Input datasets:")
        for path in datasets_paths:
            dataset = cls(path)
            dataset.summary()
            classes = dataset.classes
            reindex_dict = cls._get_reindex_dict(classes, merged_classes)

            for subset_path in [
                dataset.train_folder,
                dataset.valid_folder,
                dataset.test_folder,
            ]:
                if subset_path.exists():
                    target_subset_path = merged_dataset_path / subset_path.name
                    create_folder(target_subset_path / "images")
                    create_folder(target_subset_path / "labels")
                    images_paths = list_files(str(subset_path / "images"))
                    for image_path in images_paths:
                        yolo_image = YoloImage(image_path)
                        yolo_image.copy_to(
                            target_subset_path,
                            dataset.name + "_",
                            reindex_dict,
                        )

        cls._create_simple_yaml_file(str(merged_dataset_path), merged_classes)

        print("Merged dataset:")
        dataset = cls(str(merged_dataset_path))
        dataset.summary()

    def split(self, splitted_dataset_path, ratios=[0.9]):
        """
        Split the dataset into train, validation, and test subsets.

        Returns:
            splitted_dataset_path (str): Path to the folder where the splitted dataset will be saved.
            ratios (list[float]): Ratios for train, validation, and test subsets. Default is [0.9].
        """
        self.summary()

        images_paths = list()
        for folder_path in [
            self.train_folder,
            self.valid_folder,
            self.test_folder,
        ]:
            if folder_path.exists():
                images_paths += list_files(str(folder_path / "images"))

        random.shuffle(images_paths)

        total_paths = len(images_paths)

        train_count = int(total_paths * ratios[0])
        train_paths = images_paths[:train_count]

        if len(ratios) == 1:
            valid_paths = images_paths[train_count:]
            test_paths = list()
        else:
            valid_count = int(total_paths * ratios[1])
            valid_paths = images_paths[train_count : train_count + valid_count]
            test_paths = images_paths[train_count + valid_count :]

        splitted_dataset_path = Path(splitted_dataset_path)

        subsets = [
            ("train", train_paths),
            ("valid", valid_paths),
            ("test", test_paths),
        ]
        for subset_name, paths in subsets:
            for path in paths:
                yolo_image = YoloImage(path)
                old_subset_name = Path(path).parent.parent.name
                target_folder_path = splitted_dataset_path / subset_name
                create_folder(target_folder_path / "images")
                create_folder(target_folder_path / "labels")
                yolo_image.copy_to(
                    target_folder_path, f"{self.name}_{old_subset_name}_"
                )

        self._create_simple_yaml_file(str(splitted_dataset_path), self.classes)
        YoloDataset(str(splitted_dataset_path)).summary()

    def rename_classes(self, renamed_dataset_path, renaming_dict):
        """
        Rename classes in the dataset.

        Returns:
            renamed_dataset_path (str): Path to the folder where the renamed dataset will be saved.
            renaming_dict (dict): Dictionary mapping original class names to new class names.
        """

        self.summary()

        target_classes = [
            class_name
            for class_name in renaming_dict.values()
            if class_name is not None
        ]

        for class_name in self.classes:
            if class_name not in renaming_dict.keys():
                target_classes.append(class_name)

        target_classes = list(set(target_classes))
        target_classes.sort()

        print("Classes after renaming:", target_classes)

        reindex_dict = dict()

        for i, class_name in enumerate(self.classes):
            renamed_class = class_name
            if class_name in renaming_dict:
                renamed_class = renaming_dict[class_name]

            if renamed_class is None:
                reindex_dict[i] = None
            else:
                reindex_dict[i] = target_classes.index(renamed_class)

        renamed_dataset_path = Path(renamed_dataset_path)
        for subset_name, folder_path in [
            ("train", self.train_folder),
            ("valid", self.valid_folder),
            ("test", self.test_folder),
        ]:
            if folder_path.exists():
                images_paths = list_files(folder_path / "images")
                for path in images_paths:
                    yolo_image = YoloImage(path)
                    target_folder_path = renamed_dataset_path / subset_name
                    create_folder(target_folder_path / "images")
                    create_folder(target_folder_path / "labels")
                    yolo_image.copy_to(
                        target_folder_path, f"{self.name}_", reindex_dict
                    )

        self._create_simple_yaml_file(
            str(renamed_dataset_path), target_classes
        )
        YoloDataset(str(renamed_dataset_path)).summary()

    def show_sample(self, figsize=(10, 10)):
        """
        Show a sample of annotated images from the dataset.

        Returns:
            figsize (tuple(int, int)): Size of the figure (width, height) in inches.
        """

        images_paths = list()
        for folder_path in [
            self.train_folder,
            self.valid_folder,
            self.test_folder,
        ]:
            if folder_path.exists():
                images_paths += list_files(str(folder_path / "images"))

        random.shuffle(images_paths)

        labels_dict = {
            i: class_name for i, class_name in enumerate(self.classes)
        }
        annotated_images = list()
        if len(images_paths) > 9:
            for i in range(3):
                annotated_images.append(list())
                for j in range(3):
                    yolo_image = YoloImage(
                        images_paths[i * 3 + j], labels_dict, self.task
                    )
                    annotated_images[i].append(
                        yolo_image.get_annotated_image()
                    )
        else:
            total_images = min(3, len(images_paths))
            for i in range(total_images):
                yolo_image = YoloImage(
                    images_paths[i * 3 + j], labels_dict, self.task
                )
                annotated_images.append(yolo_image.get_annotated_image())

        concatenated_images = concatenate_images(annotated_images)

        show_image(concatenated_images, figsize=figsize)

    def number_filenames(self, numbered_dataset_path):
        """
        Number filenames of images in the dataset.

        Returns:
            numbered_dataset_path (str): Path to the folder where the numbered dataset will be saved.
        """
        self.summary()

        total_images = 0

        for folder_path in [
            self.train_folder,
            self.valid_folder,
            self.test_folder,
        ]:
            if folder_path.exists():
                total_images += len(list_files(str(folder_path / "images")))

        n_digits = len(str(total_images))
        i = 0

        subsets = [
            ("train", self.train_folder),
            ("valid", self.valid_folder),
            ("test", self.test_folder),
        ]
        numbered_dataset_path = Path(numbered_dataset_path)
        for subset_name, folder_path in subsets:
            if folder_path.exists():
                images_paths = list_files(str(folder_path / "images"))
                for path in images_paths:
                    yolo_image = YoloImage(path)
                    target_folder_path = numbered_dataset_path / subset_name
                    create_folder(target_folder_path / "images")
                    create_folder(target_folder_path / "labels")
                    yolo_image.copy_to(
                        target_folder_path, target_stem=str(i).zfill(n_digits)
                    )
                    i += 1

        self._create_simple_yaml_file(str(numbered_dataset_path), self.classes)
        YoloDataset(str(numbered_dataset_path)).summary()



import os
import re
import shutil
import subprocess
import uuid
import zipfile
from pathlib import Path

VERSION_OPERATORS = [
    "==",
    "===",
    "~=",
    "!=",
    "<",
    "<=",
    ">",
    ">=",
    # 'and', 'or', 'in', 'not in',
]

IGNORE_START_WITH = ('"', "#", "-", "git+")


def generate_random_filename():
    random_uuid = uuid.uuid4()
    return str(random_uuid)


def get_size(path, unit="bytes"):
    path = Path(path)

    if path.is_file():
        size = path.stat().st_size
    elif path.is_dir():
        total_size = 0
        for sub_path in path.glob("**/*"):
            if sub_path.is_file():
                total_size += sub_path.stat().st_size
        size = total_size
    else:
        raise ValueError("Path is neither a file nor a directory.")

    if unit == "bytes":
        return size
    elif unit == "KB":
        return size / 1024
    elif unit == "MB":
        return size / (1024**2)
    elif unit == "GB":
        return size / (1024**3)
    else:
        raise ValueError("Invalid unit")


def list_files(folder_path, return_absolute_paths=True):
    """
    List files in a folder.

    Args:
        folder_path (str): Path to the folder.
        return_absolute_paths (bool): Whether to return absolute paths or relative paths.

    Returns:
        sorted_files_paths (list): A list of file paths.
    """
    folder_path = Path(folder_path)
    items = list(folder_path.iterdir())

    files = [item.resolve() for item in items if item.is_file()]
    sorted_files = sorted(files)

    if not return_absolute_paths:
        sorted_files = [file.relative_to(folder_path) for file in sorted_files]

    sorted_files_paths = [str(file.as_posix()) for file in sorted_files]

    return sorted_files_paths


def list_folders(folder_path, return_absolute_paths=True):
    """
    List folders in a folder.

    Args:
        folder_path (str): Path to the folder.
        return_absolute_paths (bool): Whether to return absolute paths or relative paths.

    Returns:
        sorted_folders_paths (list): A list of folder paths.
    """
    folder_path = Path(folder_path)
    items = list(folder_path.iterdir())

    folders = [item.resolve() for item in items if item.is_dir()]

    sorted_folders = sorted(folders)

    if not return_absolute_paths:
        sorted_folders = [
            folder.relative_to(folder_path) for folder in sorted_folders
        ]

    sorted_folders_paths = [str(file.as_posix()) for file in sorted_folders]

    return sorted_folders_paths


def list_contents(folder_path, return_absolute_paths=True):
    return list_files(folder_path, return_absolute_paths), list_folders(
        folder_path, return_absolute_paths
    )


def get_folder_details(folder_path):
    files_paths, folders_paths = list_contents(folder_path)
    files_count, folders_count = len(files_paths), len(folders_paths)
    return f"({files_count} files + {folders_count} folders)"


def print_foldertree(folder_path, level=0, max_level=1, verbose=True):
    """
    Print the folder tree structure.

    Args:
        folder_path (str): Path to the folder.
        level (int): Current level to apply recursion, you do not need to provide this field.
        max_level (int): Maximum level of folder tree.
        verbose (bool): Whether to print additional details.
    """
    if level == 0:
        current_line = f"{folder_path} "
        if verbose:
            current_line += get_folder_details(folder_path)
        print(current_line)

    files_paths, child_folders_paths = list_contents(folder_path)

    if len(files_paths) > 0:
        print("|   " * (level + 1))

    for file_path in files_paths:
        file_name = Path(file_path).name
        current_line = "|   " * level + f"|-- {file_name} "
        if verbose:
            file_size = get_size(str(file_path), "KB")
            current_line += f"({file_size:.2f} KB)"
        print(current_line)

    if len(child_folders_paths) > 0:
        print("|   " * (level + 1))

    for i, child_folder_path in enumerate(child_folders_paths):
        child_folder_name = Path(child_folder_path).name
        current_line = "|   " * level + f"|-- {child_folder_name} "
        if verbose:
            current_line += get_folder_details(child_folder_path)
        print(current_line)
        if level + 1 < max_level:
            print_foldertree(child_folder_path, level + 1, verbose)

        if i + 1 < len(child_folders_paths):
            print("|   " * (level + 1))


def zip_paths(paths, output_zip):
    """
    Create a ZIP file containing specified paths.

    Args:
        paths (list): List of file or folder paths to be included in the ZIP file.
        output_zip (str): Path to save the output ZIP file.
    """
    with zipfile.ZipFile(output_zip, "w") as zipf:
        for path in paths:
            if os.path.isdir(path):
                folder_path = path
                for folder_root, _, files in os.walk(folder_path):
                    for file in files:
                        file_path = os.path.join(folder_root, file)
                        arcname = os.path.relpath(file_path, folder_path)
                        zipf.write(
                            file_path,
                            arcname=os.path.join(
                                os.path.basename(folder_path), arcname
                            ),
                        )
            elif os.path.isfile(path):
                file_path = path
                filename = os.path.basename(file_path)
                zipf.write(file_path, arcname=filename)


def remove_folder(folder_path):
    """
    Remove a folder and its contents.

    Args:
        folder_path (str): Path to the folder to be removed.
    """
    try:
        shutil.rmtree(folder_path)
        print(f"Folder '{folder_path}' and its contents removed successfully.")
    except Exception as e:
        print(f"An error occurred: {e}")


def create_folder(folder_path):
    """
    Create a folder if it doesn't exist.

    Args:
        folder_path (str): Path to the folder to be created.
    """
    os.makedirs(folder_path, exist_ok=True)


def split_file_path(file_path):
    folders = []
    path = file_path
    while True:
        path, folder = os.path.split(path)
        if folder:
            folders.insert(0, folder)
        else:
            break

    return folders


def get_package_names(requirements_file_path):
    package_names = []
    with open(requirements_file_path, "r") as file:
        for line in file:
            line = line.strip()
            if line and not line.startswith(IGNORE_START_WITH):
                package_name = line
                for op in VERSION_OPERATORS:
                    if op in line:
                        package_name = re.split(
                            f"\s*{re.escape(op)}\s*", package_name, maxsplit=1
                        )[0]

                package_names.append(package_name.strip())

    return package_names


def get_installed_versions(package_names):
    installed_versions = {}
    for package_name in package_names:
        result = subprocess.run(
            ["pip", "show", package_name], capture_output=True, text=True
        )
        output = result.stdout.strip()
        lines = output.split("\n")
        version_line = [line for line in lines if line.startswith("Version:")]

        if version_line:
            version = version_line[0].split(": ")[1].strip()
            installed_versions[package_name] = version
        else:
            installed_versions[package_name] = None

    return installed_versions


def print_package_versions(requirements_file_path="requirements.txt"):
    """
    Print packages listed in requirements file and their version in the current environment

    Args:
        requirements_file_path (str): path to a file contain packages names
    """
    package_names = get_package_names(requirements_file_path)
    installed_versions = get_installed_versions(package_names)

    for package_name, version in installed_versions.items():
        if version is None:
            print(f"# {package_name}")
        else:
            print(f"{package_name}=={version}")



import cv2


def draw_skeleton(image, keypoints):
    if isinstance(image, str):
        full_image = cv2.imread(image)
    else:
        full_image = image.copy()

    # draw points
    head_indices = [0, 1, 2, 3, 4]
    body_indices = [5, 6, 7, 8, 9, 10]
    # leg_indices = [11, 12, 13, 14, 15, 16]

    for i, keypoint in enumerate(keypoints):
        if i in head_indices:
            color = (255, 0, 0)
        elif i in body_indices:
            color = (0, 0, 255)
        else:
            color = (0, 255, 0)

        xy = keypoint["xy"]
        state = keypoint["state"]

        if state == 1:
            thickness = 2
        else:
            thickness = -1

        if xy != [0, 0] and state != 0:
            cv2.circle(
                full_image, xy, radius=10, color=color, thickness=thickness
            )

    # draw lines
    head_lines = [6, 4, 2, 0, 1, 3, 5]
    body_lines = [10, 8, 6, 5, 7, 9]
    leg_lines = [16, 14, 12, 11, 13, 15]

    line_thickness = 2

    for id1, id2 in zip(head_lines[:-1], head_lines[1:]):
        if keypoints[id1]["state"] != 0 and keypoints[id2]["state"] != 0:
            cv2.line(
                full_image,
                keypoints[id1]["xy"],
                keypoints[id2]["xy"],
                (255, 0, 0),
                line_thickness,
            )

    for id1, id2 in zip(body_lines[:-1], body_lines[1:]):
        if keypoints[id1]["state"] != 0 and keypoints[id2]["state"] != 0:
            cv2.line(
                full_image,
                keypoints[id1]["xy"],
                keypoints[id2]["xy"],
                (0, 0, 255),
                line_thickness,
            )

    if keypoints[12]["state"] != 0 and keypoints[6]["state"] != 0:
        cv2.line(
            full_image,
            keypoints[12]["xy"],
            keypoints[6]["xy"],
            (0, 0, 255),
            line_thickness,
        )
    if keypoints[11]["state"] != 0 and keypoints[5]["state"] != 0:
        cv2.line(
            full_image,
            keypoints[11]["xy"],
            keypoints[5]["xy"],
            (0, 0, 255),
            line_thickness,
        )

    for id1, id2 in zip(leg_lines[:-1], leg_lines[1:]):
        if keypoints[id1]["state"] != 0 and keypoints[id2]["state"] != 0:
            cv2.line(
                full_image,
                keypoints[id1]["xy"],
                keypoints[id2]["xy"],
                (0, 255, 0),
                line_thickness,
            )

    return full_image



import copy
from io import BytesIO

import cv2
import matplotlib.pyplot as plt
import numpy as np
import requests


def show_image(image, figsize=(10, 10)):
    """
    Show image from a numpy array or a file path.
    Which can be used when run .py or .ipynb files.

    Args:
        image (Union[np.ndarray, str]): a numpy array or a file path
        figsize (Tuple[int, int]): (width, height) for image to show
    """
    if isinstance(image, str):
        temp_image = cv2.imread(image)
    else:
        temp_image = image.copy()

    plt.figure(figsize=figsize)
    temp_image = cv2.cvtColor(temp_image.astype(np.uint8), cv2.COLOR_BGR2RGB)
    plt.imshow(temp_image)
    plt.show()


def save_image(image, save_path):
    cv2.imwrite(save_path, image)


def download_image(url, save_path=None):
    """
    Download image from given url

    Args:
        url (str): url of the image
        save_path (str): path to save image

    Returns:
        image (Union[np.ndarray, NoneType]): return numpy array of the image if
            it's downloaded successfullly. Otherwise return None
    """
    response = requests.get(url)
    if response.status_code == 200:
        image_stream = BytesIO(response.content)
        image_data = np.frombuffer(image_stream.read(), dtype=np.uint8)
        image = cv2.imdecode(image_data, cv2.IMREAD_COLOR)

        if save_path:
            cv2.imwrite(save_path, image)

        return image
    else:
        print(f"Failed to download image from {url}")
        return None


def get_random_image(width=400, height=300, save_path=None):
    """
    Download a random image with desired size

    Args:
        width (int): desired width
        height (int): desired height
        save_path (str): path to save image

    Returns:
        image (np.ndarray): numpy array of the downloaded image
    """
    image = download_image(
        f"https://picsum.photos/{width}/{height}", save_path
    )
    return image


def rotate_image(image, degree, expand=False):
    """
    Rotate the given image with given degree

    Args:
        image (np.ndarray): numpy array of the image
        degree (float): total degrees to rotate
        expand (bool): Pad rotated image with black color if True,
            otherwise return the cropped rotated image

    Returns:
        rotated_image (np.ndarray): numpy array of the rotated image
    """
    height, width = image.shape[:2]
    center_point = (width / 2, height / 2)

    rotation_matrix = cv2.getRotationMatrix2D(center_point, degree, 1.0)

    new_height, new_width = image.shape[:2]
    if expand:
        abs_cos = abs(rotation_matrix[0, 0])
        abs_sin = abs(rotation_matrix[0, 1])

        new_width = int(height * abs_sin + width * abs_cos)
        new_height = int(height * abs_cos + width * abs_sin)

        rotation_matrix[0, 2] += new_width / 2 - center_point[0]
        rotation_matrix[1, 2] += new_height / 2 - center_point[1]

    rotated_image = cv2.warpAffine(
        image, rotation_matrix, (new_width, new_height)
    )

    return rotated_image


def resize_with_black_padding(image, new_height, new_width):
    original_height, original_width = image.shape[:2]

    if len(image.shape) == 3:
        padded_image = np.zeros(
            (new_height, new_width, image.shape[2]), dtype=np.uint8
        )
        padded_image[:original_height, :original_width, :] = image
    else:
        padded_image = np.zeros((new_height, new_width), dtype=np.uint8)
        padded_image[:original_height, :original_width] = image

    return padded_image


def shift_image(image, dx, dy, new_height=None, new_width=None):
    padded_image = image.copy()

    if new_height and new_width:
        padded_image = resize_with_black_padding(
            padded_image, new_height, new_width
        )

    translation_matrix = np.float32([[1, 0, dx], [0, 1, dy]])
    shifted_image = cv2.warpAffine(
        padded_image,
        translation_matrix,
        (padded_image.shape[1], padded_image.shape[0]),
    )
    return shifted_image


def pad_image(image, target_size):
    # target_size = (height, width)
    height, width = image.shape[:2]

    pad_height = max(0, target_size[0] - height)
    pad_width = max(0, target_size[1] - width)

    top_pad = pad_height // 2
    bottom_pad = pad_height - top_pad
    left_pad = pad_width // 2
    right_pad = pad_width - left_pad

    padded_image = cv2.copyMakeBorder(
        image.copy(),
        top_pad,
        bottom_pad,
        left_pad,
        right_pad,
        cv2.BORDER_CONSTANT,
        value=(0, 0, 0),
    )

    return padded_image


def concatenate_images(image_list, padding_size=0):
    """
    Concatenate images based on its appearance order in the given 2D list
    Each image will be padded to the max height, max width in the given list.

    Args:
        image_list (list(list(np.ndarray))): 2D (or 1D) list of images
        padding_size (int): padding distance between each image

    Returns:
        concatenated_image (np.ndarray): the concatenated image from 2D list
    """

    # ensure image_list is a 2D list
    new_image_list = copy.deepcopy(image_list)
    if not isinstance(image_list[0], list):
        new_image_list = [copy.deepcopy(image_list)]

    rows = len(new_image_list)
    cols = max(len(row) for row in new_image_list)

    for image_row in new_image_list:
        image_row += [None] * (cols - len(image_row))

    # pad every images
    max_height = max(
        image.shape[0]
        for row in new_image_list
        for image in row
        if image is not None
    )
    max_width = max(
        image.shape[1]
        for row in new_image_list
        for image in row
        if image is not None
    )

    padded_image_list = list()
    for i, row in enumerate(new_image_list):
        padded_image_list.append(list())
        for j, image in enumerate(row):
            padded_image = np.zeros((max_height, max_width, 3), dtype=np.uint8)
            if image is not None:
                padded_image = pad_image(image, (max_height, max_width))
            padded_image_list[i].append(padded_image)

    concatenated_image = np.zeros(
        (
            max_height * rows + padding_size * (rows - 1),
            max_width * cols + padding_size * (cols - 1),
            3,
        ),
        dtype=np.uint8,
    )

    for i, row in enumerate(padded_image_list):
        for j, image in enumerate(row):

            h, w, _ = image.shape
            concatenated_image[
                (i > 0) * padding_size
                + i * max_height : (i > 0) * padding_size
                + i * max_height
                + h,
                (j > 0) * padding_size
                + j * max_width : (j > 0) * padding_size
                + j * max_width
                + w,
            ] = image

    return concatenated_image
